from statefun_tasks import FlinkTasks
import asyncio
import cloudpickle
import logging


_log = logging.getLogger('FlinkTasks')
__defaults = None


def enable_inline_tasks(tasks: FlinkTasks):
    """
    Enables inline tasks support.  Inline tasks work by sending pickled code as well as data 
    to a general purpose run_code task as an alternative to deploying functions decorated with 
    @tasks.bind()

    N.B. there are security considerations whenever using pickled code.  Only accept code from 
    trusted sources and consider implementing your own implementation of this extension with suitable
    checks - e.g. you might want to only accept signed code or restrict what global functions and 
    imports are available.  This extension is included as an example of what is possible.

    :param tasks: FlinkTasks to enable @inline_task() for
    """
    @tasks.bind(module_name='__builtins', with_context=True, with_state=True)
    async def run_code(context, state, __with_context, __with_state, __code, *args, **kwargs):
        
        fn = cloudpickle.loads(__code)
        fn_args = []

        if __with_context:
            fn_args.append(context)

        if __with_state:
            fn_args.append(state)

        fn_args.extend(args)

        safer_locals = {'fn': fn, 'args': fn_args, 'kwargs': kwargs}

        # sample restrictions on loaded code
        # ----------------------------------
        #
        # safer_builtins = {**fn.__globals__['__builtins__']}
        # fn.__globals__['__builtins__'] = safer_builtins
        # del safer_builtins['exit']

        # safer_open = open
        # def safer_open(file, *sargs, **skwargs):
        #     # you might check and raise error if file not in valid list of files for example...
        #     open(file, *sargs, **skwargs)
        # safer_builtins['open'] = safer_open

        exec('__res = fn(*args, **kwargs)', {}, safer_locals)
        res = safer_locals['__res']

        if asyncio.iscoroutinefunction(fn):
            res = await res

        if __with_state:
            return res
        else:
            return state, res

    global __defaults
    __defaults = run_code.defaults()

    _log.warning('Inline tasks enabled. This is a potential security risk')


def inline_task(include=None, depends=None, with_context=False, with_state=False, **params):
    """
    Declares an inline Flink task
    :param include: list of modules to include with the pickled code
    :param depends: list of tasks this tasks depends on (i.e. will invoke as a pipeline).  Includes from these tasks will be added automatically
    :param with_context: If set the first parameter to the function is exepcted to be the task context
    :param with_state: If set the next parameter is expected to be the task state
    :param params: any additional parameters to the Flink Task (such as a retry policy)
    :return: inline Flink task
    """
    
    includes = include or []
    depends = depends or []

    def _includes():
        all_includes = set(includes)

        for dependency in depends:
            all_includes.update(dependency.includes())

        return list(all_includes)


    def pickle(fn):
        modules = _includes()

        for module in modules:
            try:
                cloudpickle.register_pickle_by_value(module)
            except ValueError:
                # ignore cloudpickle errors about modules not yet being imported
                pass

        code = cloudpickle.dumps(fn)

        for module in modules:
            try:
                cloudpickle.unregister_pickle_by_value(module)
            except ValueError:
                # ignore cloudpickle errors about modules not yet being imported
                pass

        return code

    def decorator(fn):

        params.setdefault('display_name', f'{fn.__module__}.{fn.__name__}')

        def send(*args, **kwargs):

            if __defaults is None:
                raise ValueError('Inline tasks should be enabled with enable_inline_tasks() first')

            code = pickle(fn)

            def run_code():
                pass

            fn_kwargs = {**kwargs, '__with_context': with_context, '__with_state': with_state, '__code': code}
            task_params = {**__defaults, **params}
            return FlinkTasks.extend(run_code, **task_params).send(*args, **fn_kwargs)

        def to_task(args, kwargs, is_finally=False, parameters=None):

            if __defaults is None:
                raise ValueError('Inline tasks should be enabled with enable_inline_tasks() first')

            code = pickle(fn)

            def run_code():
                pass

            fn_kwargs = {**kwargs, '__with_context': with_context, '__with_state': with_state, '__code': code}
            task_params = {**__defaults, **params}
            return FlinkTasks.extend(run_code, **task_params).to_task(args, fn_kwargs, is_finally, {})

        fn.includes = _includes
        fn.send = send
        fn.to_task = to_task

        return fn
    
    return decorator



from re import L
from statefun_tasks.storage import StorageBackend
from statefun_tasks.serialisation import DefaultSerialiser
from statefun_tasks.pipeline_builder import PipelineBuilder

from statefun_tasks.types import Task, RetryPolicy, \
    TASK_STATE_TYPE, TASK_REQUEST_TYPE, TASK_RESULT_TYPE, TASK_EXCEPTION_TYPE, PIPELINE_STATE_TYPE

from statefun_tasks.messages_pb2 import TaskResult, TaskException, TaskRequest, Address
from statefun_tasks.type_helpers import _create_task_result, _create_task_exception
from statefun_tasks.context import TaskContext
from statefun_tasks.utils import _task_type_for, _unpack_single_tuple_args, _gen_id
from statefun_tasks.pipeline import _Pipeline
from statefun_tasks.tasks import FlinkTask
from statefun_tasks.task_impl.handlers import TaskRequestHandler, TaskResponseHandler, TaskActionHandler, ChildPipelineHandler
from statefun_tasks.events import EventHandlers

from statefun import ValueSpec, Context, Message
import logging

_log = logging.getLogger('FlinkTasks')


def _run_pipeline(pipeline_proto):
    return PipelineBuilder.from_proto(pipeline_proto).validate()


class FlinkTasks(object):
    """
    Flink Tasks implementation

    An instance should be instantiated at the top of your api file and used to decorate functions
    to be exposed as tasks.

    :param default_namespace: namespace to expose functions under. Maps to Flink Statefun function namespace in module.yaml
    :param default_worker_name: worker name to expose.  Maps to Flink Statefun function type in module.yaml
    :param egress_type_name: egress type name.  Maps to Flink Statefun egress in module.yaml
    :param optional serialiser: serialiser to use (will use DefaultSerialiser if not set)
    """
    def __init__(self, default_namespace: str = None, default_worker_name: str = None, egress_type_name: str = None, serialiser = None):
        self._default_namespace = default_namespace
        self._default_worker_name = default_worker_name
        self._egress_type_name = egress_type_name
        self._serialiser = serialiser if serialiser is not None else DefaultSerialiser()
        self._bindings = {}
        self._events = EventHandlers()
        self._storage = None
        
        # to register in Flink's @functions.bind() attribute
        self._value_specs = [
            ValueSpec(name="task_request", type=TASK_REQUEST_TYPE),
            ValueSpec(name="task_result", type=TASK_RESULT_TYPE),
            ValueSpec(name="task_exception", type=TASK_EXCEPTION_TYPE),
            ValueSpec(name="task_state", type=TASK_STATE_TYPE),
            ValueSpec(name="pipeline_state", type=PIPELINE_STATE_TYPE),
        ]

        self.register_builtin('run_pipeline', _run_pipeline)

        self._handlers = [
            TaskRequestHandler(),
            TaskResponseHandler(),
            TaskActionHandler(),
            ChildPipelineHandler()
        ]

    @staticmethod
    def extend(function, retry_policy: RetryPolicy = None, **params):
        """
        Adds the extensions to make a function unsable by a PipelineBuilder -> fn.send(), fn.to_task() and fn.defaults()

        :param function: the function to wrap
        :param retry_policy: retry policy to use should the task throw an exception
        :param params: any additional parameters to the Flink Task (such as a retry policy)
        """
        def defaults():
            return {
                'retry_policy': None if retry_policy is None else retry_policy.to_proto(),
                **params
            }

        def send(*args, **kwargs):
            return PipelineBuilder().send(function, *args, **kwargs)

        def to_task(args, kwargs, is_finally=False, parameters={}):
            parameters = {**defaults(), **parameters}

            if is_finally:
                parameters['is_fruitful'] = False
            
            module_name = parameters.get('module_name', None)
            task_type = _task_type_for(function, module_name)

            task_id = parameters.pop('task_id') or _gen_id()
            args = _unpack_single_tuple_args(args)

            return Task.from_fields(task_id, task_type, args, kwargs, is_finally=is_finally, **parameters)

        function.send = send
        function.to_task = to_task
        function.defaults = defaults

        return function

    def bind(
        self, 
        namespace: str = None, 
        worker_name: str = None, 
        retry_policy: RetryPolicy = None, 
        with_state: bool = False, 
        is_fruitful: bool = True, 
        module_name: str = None,
        with_context: bool = False,
        display_name: str = None,
        task_id: str = None):
        """
        Decorator to bind a function as a Flink Task

        :param namespace: namespace to use in place of the default
        :param worker_name: worker name to use in place of the default
        :param retry_policy: retry policy to use should the task throw an exception
        :param with_state: whether to pass a state object as the first (second if with_context is also set) parameter.  The return value should be a tuple of state, result (default False)
        :param is_fruitful: whether the function produces a fruitful result or simply returns None (default True)
        :param module_name: if specified then the task type used in addressing will be module_name.function_name
                            otherwise the Python module containing the function will be used
        :param with_context: whether to pass a Flink context object as the first parameter (default false)
        :param display_name: optional friendly name for this task
        :param task_id: optional set the fixed id for this task in order to make it stateful
        """
        
        def wrapper(function):

            function = FlinkTasks.extend(function, 
                    namespace = self._default_namespace if namespace is None else namespace,
                    worker_name = self._default_worker_name if worker_name is None else worker_name,
                    retry_policy = retry_policy,
                    with_state = with_state,
                    is_fruitful = is_fruitful,
                    module_name = module_name,
                    with_context = with_context,
                    display_name = display_name,
                    task_id = task_id
            )

            self.register(function, wrapper=None, **function.defaults())
            return function

        return wrapper

    @property
    def events(self) -> EventHandlers:
        """
        EventHandler for this FlinkTasks instance
        """
        return self._events

    def set_storage_backend(self, storage: StorageBackend):
        """
        Sets the storage backend to use if required. The default is to use Flink state exclusively but temporary results from large parallelisms 
        may cause memory issues in which case they can be saved outside in e.g. an S3 bucket or Redis

        :param storage: results storage backend to use
        """
        self._storage = storage

    def value_specs(self):
        return self._value_specs

    def register_builtin(self, type_name, fun, **params):
        """
        Registers a built in Flink Task e.g. __builtins.run_pipeline
        This should only be used if you want to provide your own built in (pre-defined) tasks

        :param type_name: task type to expose the built in as
        :param fun: a function, partial or lambda representing the built in
        :param params: any additional parameters to the Flink Task (such as a retry policy)
        """
        self._bindings[f'__builtins.{type_name}'] = FlinkTask(fun, self._serialiser, self.events, **params)

    def register(self, fun, wrapper=None, module_name=None, **params):
        """
        Registers a Python function as a Flink Task.
        Normally you would attribute the function with @tasks.bind() instead

        :param fun: the python function
        :param optional wrapper: if wrapping a task function with e.g. functools.wraps then pass the wrapper here
        :param optional module_name: the module name to register the task under which by default is the Python module name containing the function
        :param params: any additional parameters to the Flink Task (such as a retry policy)
        """

        if fun is None:
            raise ValueError("function instance must be provided")

        fun.type_name = _task_type_for(fun, module_name)
        self._bindings[fun.type_name] = FlinkTask(wrapper or fun, self._serialiser, self.events, **params)


    def get_task(self, task_type) -> FlinkTask:
        """
        Returns the Flink Task instance for a given task type name

        :param task_type: task type name e.g. examples.multiply
        :return: the Flink Task
        """
        if task_type in self._bindings:
            return self._bindings[task_type]
        else:
            raise RuntimeError(f'{task_type} is not a registered FlinkTask')

    async def run_async(self, context: Context, message: Message):
        """
        Runs an async Flink Task

        :param context: context object provided by Flink
        :param message: the task input protobuf message
        """
        with TaskContext(context, self._egress_type_name, self._serialiser) as task_context:

            for handler in self._handlers:
                task_input = handler.unpack(task_context, message)
                if task_input is not None:
                    
                    try:
                        _log.info(f'Starting {task_context}')

                        await handler.handle_message(self, task_context, task_input)

                        _log.info(f'Finished {task_context}')

                    except Exception as ex:
                        _log.error(f'Error invoking {task_context} - {ex}')
                        self.fail(task_context, task_input, ex)
                        
                    finally:
                        return
            
            _log.error(f'Unsupported message type {message.typed_value.typename}')



    @staticmethod
    def send(func, *args, **kwargs) -> PipelineBuilder:
        """
        Returns a PipelineBuilder with this task as the first item in the pipeline

        :param func: a function decorated with @tasks.bind()
        :param args: task args
        :param kwargs: task kwargs
        :return: a pipeline builder
        """
        try:
            send_func = func.send
        except AttributeError:
            raise AttributeError(
                'Expected function to have a send attribute. Make sure it is decorated with @tasks.bind()')
        return send_func(*args, **kwargs)

    def clone_task_request(self, context: TaskContext) -> TaskRequest:
        """
        Returns a complete copy of the TaskRequest associated with this TaskContext with the 
        caller address details extracted as necessary from the context

        :param context: TaskContext
        :return: a TaskRequest
        """
        task_request = TaskRequest()
        task_request.CopyFrom(context.storage.task_request)

        # copy caller details from context unless explictly set already
        if not task_request.HasField('reply_address') and not task_request.HasField('reply_topic'):
            caller = context.get_caller()

            if caller is not None:
                task_request.reply_address.CopyFrom(caller)

        return task_request

    def unpack_task_request(self, task_request: TaskRequest) -> tuple:
        """
        Unpacks a TaskRequest into args, kwargs and state

        :param task_request: TaskRequest
        :return: args, kwargs and state from this task_request
        """
        args, kwargs, state = self._serialiser.deserialise_request(task_request)
        return args, kwargs, state

    def send_result(self, context: TaskContext, task_request: TaskRequest, result, state=Ellipsis):
        """
        Sends a result
        :param context: TaskContext
        :param task_request: the incoming TaskRequest
        :param result: the result(s) to return
        :param state: the state to include in the result.  If not specified it will be copied from the TaskRequest
        """
        task_result = _create_task_result(task_request)

        if state == Ellipsis:  # default value since None is perfectly valid state
            state = task_request.state

        self._serialiser.serialise_result(task_result, result, state)

        self.emit_result(context, task_request, task_result)

    def fail(self, context, task_input, ex):
        """
        Sends a failure
        :param context: TaskContext
        :param task_input: the incoming TaskRequest or TaskActionRequest
        :param ex: the exception to return
        """
        task_exception = _create_task_exception(task_input, ex)
        self.emit_result(context, task_input, task_exception)

    def get_pipeline(self, context):
        pipeline_protos = context.pipeline_state.pipeline

        if pipeline_protos is not None:
            return _Pipeline.from_proto(pipeline_protos, self._serialiser, self._events, self._storage)
        else:
            raise ValueError(f'Missing pipeline for task_id - {context.get_task_id()}')

    def try_get_pipeline(self, context):
        try:
            return self.get_pipeline(context)
        except:
            return None

    def emit_result(self, context, task_input, task_result):
        # record task result / task exception
        if isinstance(task_result, TaskResult):
            task_result.invocation_id = task_input.invocation_id
            context.storage.task_result = task_result
        
        elif isinstance(task_result, TaskException):
            task_result.invocation_id = task_input.invocation_id
            context.storage.task_exception = task_result
            
        # either send a message to egress if reply_topic was specified
        if task_input.HasField('reply_topic'):
            context.send_egress_message(topic=task_input.reply_topic, value=task_result)

        # or call back to a particular flink function if reply_address was specified
        elif task_input.HasField('reply_address'):
            address, identifer = context.to_address_and_id(task_input.reply_address)
            context.send_message(address, identifer, task_result)

        elif isinstance(task_input, TaskRequest):
            # if this was a task request it could be part of a pipeline - i.e. called from another tasks so
            # either call back to original caller (e.g. if this is a result of a retry)
            task_state = context.task_state.by_uid[task_input.uid]
            if task_state.original_caller_id != '':
                context.send_message(task_state.original_caller_address, task_state.original_caller_id, task_result)

            # or call back to our caller (if there is one)
            elif context.get_caller_id() is not None:
                context.send_message(context.get_caller_address(), context.get_caller_id(), task_result)

            # clean up
            del context.task_state.by_uid[task_input.uid]



from statefun_tasks.storage import StorageBackend
from statefun_tasks.types import Group, _type_name
from statefun_tasks.context import TaskContext
from statefun_tasks.protobuf import pack_any
from statefun_tasks.messages_pb2 import TaskResults, TaskResult, TaskException, MapOfStringToAny
from google.protobuf.wrappers_pb2 import StringValue
from google.protobuf.any_pb2 import Any
from collections import deque
import traceback as tb
import logging


_log = logging.getLogger('FlinkTasks')


class ResultAggregator(object):
    def __init__(self, graph, serialiser, storage: StorageBackend):
        self._graph = graph
        self._serialiser = serialiser
        self._storage = storage

    async def add_result(self, context: TaskContext, task_uid, task_result_or_exception):
        task_results = context.pipeline_state.task_results  # type: TaskResults

        failed = isinstance(task_result_or_exception, TaskException)
        last_task = self._graph.get_last_task_in_chain(task_uid)
        packed = pack_any(task_result_or_exception)
        
        if task_uid == last_task.uid:
            # if we are the last task in this chain in the group then record this result so we can aggregate laster
            await self._save_result(task_uid, packed, task_results, context.pipeline_state_size)

        elif failed:
            # additionally propagate the error onto the last stage of this chain
            await self._save_result(last_task.uid, packed, task_results, context.pipeline_state_size)

    async def aggregate(self, context: TaskContext, group: Group):
        is_top_level = self._graph.is_top_level_group(group)
        task_results = context.pipeline_state.task_results  # type: TaskResults

        aggregated_results, aggregated_states, aggregated_errors = [], [], []
        stack = deque([(group, aggregated_results)])  # FIFO, errors are flat not nested so don't get stacked

        while len(stack) > 0:
            group, results = stack.popleft()
            
            for pipeline in group:
                last_entry = pipeline[-1]

                if isinstance(last_entry, Group):
                    if last_entry.group_id in task_results.by_uid:
                        # if we have already aggregated this group just append the results
                        await self._append_results(last_entry.group_id, task_results, results, aggregated_states, aggregated_errors)
                    else:
                        # otherwise iterate over group
                        stack_results = []
                        results.append(stack_results)
                        stack.append((last_entry, stack_results))
                else:
                    # aggregate the group
                    await self._append_results(pipeline[-1].uid, task_results, results, aggregated_states, aggregated_errors)


        aggregated_errors = [error for error in aggregated_errors if error is not None]
        aggregated_state = self._aggregate_state(aggregated_states)

        
        if any(aggregated_errors):

            serialised_state = self._serialiser.to_proto(aggregated_state)

            task_result_or_exception = TaskException(            
                id=group.group_id,
                type=f'__aggregate.error',
                exception_type='statefun_tasks.AggregatedError',
                exception_message='|'.join([f'{e.id}, {e.type}, {e.exception_message}' for e in aggregated_errors]),
                stacktrace='|'.join([f'{e.id}, {e.stacktrace}' for e in aggregated_errors]),
                state=pack_any(serialised_state),
                invocation_id=context.pipeline_state.invocation_id)
            
        else:
        
            task_result_or_exception = TaskResult(id=group.group_id, invocation_id=context.pipeline_state.invocation_id)
            self._serialiser.serialise_result(task_result_or_exception, aggregated_results, aggregated_state)

        # save and return the aggregated result
        if not is_top_level:
            task_results.by_uid[group.group_id].CopyFrom(pack_any(task_result_or_exception))

        return task_result_or_exception

    async def _append_results(self, uid, task_results, aggregated_results, aggregated_states, aggregated_errors):
        proto = task_results.by_uid[uid]  # Any
        result, state, error = await self._load_result(proto)

        # We don't need the individual task results anymore so remove to save space / reduce message size
        del task_results.by_uid[uid] 

        aggregated_results.append(result)
        
        aggregated_states.append(state)  # states are flat not nested so don't get stacked
        aggregated_errors.append(error)  # errors are flat not nested so don't get stacked

    @staticmethod
    def _aggregate_state(aggregated_states):
        # if the group tasks all return maps then attempt to flatten the state into a single map
        # symantically equivalent to aggregated_state = dict(ChainMap(*aggregated_states)) if we were using python dicts
        if all([isinstance(state, MapOfStringToAny) for state in aggregated_states]):
            aggregated_state = MapOfStringToAny()
            for key, value in aggregated_states:
                if key not in aggregated_state.items:
                    aggregated_state.items[key] = value

        else: 
            # otherwise return state of the first task in group as we only support state mutation in groups if the state is a mergable dictionary
            aggregated_state = aggregated_states[0]

        return aggregated_state

    async def _save_result(self, uid, proto, task_results, size_of_state):
        saved_to_storage = False

        if self._storage is not None and size_of_state >= self._storage.threshold:
            saved_to_storage = await self._try_save_to_store(['aggregated_results', uid], proto)

        if saved_to_storage: 
            # record ptr to state
            task_results.by_uid[uid].CopyFrom(pack_any(StringValue(value=uid)))
        else:
            # record data to state
            task_results.by_uid[uid].CopyFrom(proto)

    async def _try_save_to_store(self, keys, proto):
        try:
            await self._storage.store(keys, proto.SerializeToString())
            return True
        except Exception as ex:
            _log.warning(f'Error saving {keys} to backend storage - {ex}')
            return False

    async def _load_result(self, proto):
        result, state, error = None, None, None
        unpacked = self._serialiser.from_proto(proto)
 
        if isinstance(unpacked, str):  # load from store
            uid = unpacked
            try:
                if self._storage is None:
                    raise ValueError('Missing storage backend')

                packed_bytes = await self._storage.fetch(['aggregated_results', uid])
                packed = Any()
                packed.ParseFromString(packed_bytes)
                unpacked = self._serialiser.from_proto(packed)

            except Exception as ex:
                _log.error(f'Error loading {uid} from backend storage - {ex}')

                unpacked = TaskException(            
                    type=f'__storage_backend.error',
                    exception_type=_type_name(ex),
                    exception_message=str(ex),
                    stacktrace=tb.format_exc())

        # now split into result, state and error tuple
        if isinstance(unpacked, TaskResult):
            result, state = unpacked.result, unpacked.state
        elif isinstance(unpacked, TaskException):
            error, state = unpacked, unpacked.state

        return result, state, error



from abc import ABC, abstractmethod
from typing import List


class StorageBackend(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def threshold(self) -> int:
        """
        Threshold in bytes at which point this storage backend is used in place of pipeline state
        """
        pass

    @abstractmethod
    async def store(self, keys: List[str], value: bytes):
        pass

    @abstractmethod
    async def fetch(self, keys: List[str]) -> bytes:
        pass



from statefun_tasks import StorageBackend, TasksException
from aiobotocore.session import get_session
from typing import List


class S3StorageBackend(StorageBackend):
    """
    An example S3 based storage backend
    :param bucket: S3 bucket name
    :param endpoint_url: endpoint url if using S3 compatible storage
    :param region_name: AWS S3 region name
    :param aws_secret_access_key: secret access key
    :param aws_access_key_id: access key id
    :param optional threshold: threshold in bytes at which this storage backend will begin offloading from pipeline state into S3
    """
    def __init__(self, bucket, endpoint_url=None, region_name=None, aws_secret_access_key=None, aws_access_key_id=None, threshold=1024000):
        self._bucket = bucket
        self._endpoint_url = endpoint_url
        self._region_name = region_name
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_access_key_id = aws_access_key_id
        self._client = None
        self._threshold = threshold

    @property
    def threshold(self) -> int:
        return self._threshold

    async def __aenter__(self):
        session = get_session()

        self._client = await session.create_client('s3', 
            endpoint_url=self._endpoint_url,
            region_name=self._region_name, 
            aws_secret_access_key=self._aws_secret_access_key,
            aws_access_key_id=self._aws_access_key_id).__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._client.__aexit__(exc_type, exc_val, exc_tb)

    async def store(self, keys: List[str], value: bytes):
        if self._client is None:
            raise TasksException('S3StorageBackend should be used with async with')

        await self._client.put_object(Bucket=self._bucket, Key='/'.join(keys), Body=value)

    async def fetch(self, keys: List[str]) -> bytes:
        if self._client is None:
            raise TasksException('S3StorageBackend should be used with async with')

        response = await self._client.get_object(Bucket=self._bucket, Key='/'.join(keys))
        async with response['Body'] as stream:
            return await stream.read()



from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TaskException, Pipeline, TaskStatus
from statefun_tasks.types import TasksException


class EventHandlers(object):
    def __init__(self):
        self._on_task_started_handlers = []
        self._on_task_finished_handlers = []
        self._on_task_retry_handlers = []
        self._on_pipeline_created_handlers = []
        self._on_pipeline_status_changed = []
        self._on_pipeline_task_finished_handlers = []
        self._on_pipeline_finished_handlers = []

    def on_task_started(self, handler):
        """
        @tasks.events.on_task_started decorator
        """
        self._on_task_started_handlers.append(handler)

    def on_task_finished(self, handler):
        """
        @tasks.events.on_task_finished decorator
        """
        self._on_task_finished_handlers.append(handler)

    def on_task_retry(self, handler):
        """
        @tasks.events.on_task_retry decorator
        """
        self._on_task_retry_handlers.append(handler)

    def on_pipeline_created(self, handler):
        """
        @tasks.events.on_pipeline_created decorator
        """
        self._on_pipeline_created_handlers.append(handler)

    def on_pipeline_status_changed(self, handler):
        """
        @tasks.events.on_pipeline_status_changed decorator
        """
        self._on_pipeline_status_changed.append(handler)

    def on_pipeline_task_finished(self, handler):
        """
        @tasks.events.on_pipeline_task_finished decorator
        """
        self._on_pipeline_task_finished_handlers.append(handler)

    def on_pipeline_finished(self, handler):
        """
        @tasks.events.on_pipeline_finished decorator
        """
        self._on_pipeline_finished_handlers.append(handler)

    def notify_task_started(self, context, task_request: TaskRequest):
        """
        Calls all on_task_started event handlers
        """
        self._notify_all(self._on_task_started_handlers, context, task_request)

    def notify_task_finished(self, context, task_result: TaskResult, task_exception: TaskException, is_pipeline: bool):
        """
        Calls all notify_task_finished event handlers
        """
        self._notify_all(self._on_task_finished_handlers, context, task_result=task_result, task_exception=task_exception, is_pipeline=is_pipeline)

    def notify_task_retry(self, context, task_request: TaskRequest, retry_count):
        """
        Calls all notify_task_retry event handlers
        """
        self._notify_all(self._on_task_retry_handlers, context, task_request, retry_count)

    def notify_pipeline_created(self, context, pipeline: Pipeline):
        """
        Calls all notify_pipeline_created event handlers
        """
        self._notify_all(self._on_pipeline_created_handlers, context, pipeline)

    def notify_pipeline_status_changed(self, context, pipeline: Pipeline, status: TaskStatus):
        """
        Calls all notify_pipeline_status_changed event handlers
        """
        self._notify_all(self._on_pipeline_status_changed, context, pipeline, status)

    def notify_pipeline_task_finished(self, context, task_result_or_exception):
        """
        Calls all notify_pipeline_task_finished event handlers
        """
        if isinstance(task_result_or_exception, TaskResult):
            task_result, task_exception = task_result_or_exception, None
        else:
            task_result, task_exception = None, task_result_or_exception

        self._notify_all(self._on_pipeline_task_finished_handlers, context, task_result=task_result, task_exception=task_exception)

    def notify_pipeline_finished(self, context, pipeline, task_result_or_exception):
        """
        Calls all notify_pipeline_finished event handlers
        """
        if isinstance(task_result_or_exception, TaskResult):
            task_result, task_exception = task_result_or_exception, None
        else:
            task_result, task_exception = None, task_result_or_exception

        self._notify_all(self._on_pipeline_finished_handlers, context, pipeline, task_result=task_result, task_exception=task_exception)


    @staticmethod
    def _notify_all(handlers, *args, **kwargs):
        for handler in handlers:
            try:
                handler(*args, **kwargs)
            except TasksException:
                raise
            except:
                pass



from statefun_tasks.context import TaskContext
from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TupleOfAny, TaskException, PausedTask, TaskStatus
from statefun_tasks.storage import StorageBackend
from statefun_tasks.types import Task
from statefun_tasks.utils import _gen_id
from typing import List
import asyncio
import logging


_log = logging.getLogger('FlinkTasks')


class DeferredTaskSubmitter(object):
    def __init__(self, graph, serialiser, storage: StorageBackend):
        self._graph = graph
        self._serialiser = serialiser
        self._storage = storage

    async def submit_tasks(self, context: TaskContext, tasks: List[Task], task_result_or_exception=None, max_parallelism=None):     
        # unpack task_result into result + state so we can merge in extras later
        if task_result_or_exception is not None:
            task_result, task_state = self._serialiser.unpack_response(task_result_or_exception)
        else:
            task_result, task_state = None, None

        # split into tasks to call vs those to defer
        if max_parallelism is None or max_parallelism < 1:
            tasks_to_call, tasks_to_defer = tasks, []
        else:
            tasks_to_call, tasks_to_defer = tasks[:max_parallelism], tasks[max_parallelism:]

        if any(tasks_to_defer):

            # create task deferral to hold these deferred tasks
            deferral_id = self._create_task_deferral(context, tasks_to_defer, task_result_or_exception)

            # parent tasks for the deferrals are the final tasks in each chain beginning with the tasks we are going to call
            # when a parent task completes, a task from the deferral will be called keep the overall parallism at max_parallelism
            for task in tasks_to_call:
                last_task = self._graph.get_last_task_in_chain(task.uid)
                context.pipeline_state.task_deferral_ids_by_task_uid[last_task.uid] = deferral_id
            
        sent_tasks = []
        for task in tasks_to_call:
            task_request = self._create_task_request(context, task, task.request, task_state, task_result, deferral=None)
            sent_tasks.append(asyncio.ensure_future(self._send_task(context, task.get_destination(), task_request)))

        if any(sent_tasks):
            await asyncio.gather(*sent_tasks)

    async def release_tasks(self, context, caller_uid, task_result_or_exception):
        
        if isinstance(task_result_or_exception, TaskException):
            parent_id = self._graph.get_last_task_in_chain(caller_uid).uid
        else:
            parent_id = caller_uid
        
        if parent_id in context.pipeline_state.task_deferral_ids_by_task_uid:
            deferral_id = context.pipeline_state.task_deferral_ids_by_task_uid[parent_id]
            task_deferral = context.pipeline_state.task_deferrals_by_id[deferral_id]

            if any(task_deferral.task_uids):
                
                task_uid = task_deferral.task_uids.pop()
                task = self._graph.get_task(task_uid) 

                task_request = self._create_deferred_task_request(context, task, task_deferral)
                
                # add task we are submitting to deferral task
                last_task = self._graph.get_last_task_in_chain(task_uid)

                context.pipeline_state.task_deferral_ids_by_task_uid[last_task.uid] = deferral_id

                # send the task
                await self._send_task(context, task.get_destination(), task_request)
            else:
                # clean up
                del context.pipeline_state.task_deferral_ids_by_task_uid[parent_id]
                del context.pipeline_state.task_deferrals_by_id[deferral_id]

    async def unpause_tasks(self, context):
        for paused_task in context.pipeline_state.paused_tasks:

            if paused_task.HasField('task_request'):
                context.send_message(paused_task.destination, paused_task.task_request.id, paused_task.task_request)
            else:
                paused_task = await self._load_from_store(context, paused_task)
                context.send_message(paused_task.destination, paused_task.task_request.id, paused_task.task_request)

        context.pipeline_state.ClearField('paused_tasks')

    def _create_task_deferral(self, context, tasks_to_defer, task_result_or_exception):
        # create task deferral to hold these deferred tasks
        deferral_id = _gen_id()
        task_deferral = context.pipeline_state.task_deferrals_by_id[deferral_id]

        if context.get_caller_id() is not None:
            task_deferral.parent_task_address = context.get_caller_address()
            task_deferral.parent_task_id = context.get_caller_id()

        if task_result_or_exception is not None:
            if isinstance(task_result_or_exception, TaskResult):
                task_deferral.task_result_or_exception.task_result.CopyFrom(task_result_or_exception)
            elif isinstance(task_result_or_exception, TaskException):
                task_deferral.task_result_or_exception.task_exception.CopyFrom(task_result_or_exception)

        for task in reversed(tasks_to_defer):
            task_deferral.task_uids.append(task.uid)

        return deferral_id

    def _create_deferred_task_request(self, context, task, task_deferral):
        task_state, task_result = None, None
        if task_deferral.HasField('task_result_or_exception'):
            
            if task_deferral.task_result_or_exception.HasField('task_result'):
                task_result_or_exception = task_deferral.task_result_or_exception.task_result
            else:
                task_result_or_exception = task_deferral.task_result_or_exception.task_exception

            task_result, task_state = self._serialiser.unpack_response(task_result_or_exception)

        return self._create_task_request(context, task, task.request, task_state, task_result, deferral=task_deferral)

    def _create_task_request(self, context, task, task_request, task_state=None, task_result=None, deferral=None):
        if task_result is not None:
            task_args_and_kwargs = self._serialiser.to_args_and_kwargs(task.request)
            task_request = self._serialiser.merge_args_and_kwargs(task_result if not task.is_finally else TupleOfAny(), task_args_and_kwargs)

        request = TaskRequest(id=task.task_id, type=task.task_type, uid=task.uid)

        # allow callers to drop the results of fruitful tasks in their pipelines if they wish
        request.is_fruitful = task.is_fruitful

        # set invocation id to pipeline invocation id
        request.invocation_id = context.pipeline_state.invocation_id

        # set extra pipeline related parameters
        request.meta['pipeline_address'] = context.pipeline_state.address
        request.meta['pipeline_id'] = context.pipeline_state.id
        request.meta['root_pipeline_id'] = context.pipeline_state.root_id
        request.meta['root_pipeline_address'] = context.pipeline_state.root_address

        if task.display_name is not None:
            request.meta['display_name'] = task.display_name
        
        if deferral is not None:  # copy parent task details from the deferral
            if deferral.HasField('parent_task_address'):
                request.meta['parent_task_address'] = deferral.parent_task_address

            if deferral.HasField('parent_task_id'):
                request.meta['parent_task_id'] = deferral.parent_task_id
        else: 
            # otherwise grab from the context
            if context.get_caller_id() is not None:
                request.meta['parent_task_address'] = context.get_caller_address()
                request.meta['parent_task_id'] = context.get_caller_id()

        self._serialiser.serialise_request(request, task_request, state=task_state, retry_policy=task.retry_policy)
        return request

    async def _send_task(self, context, destination, task_request):
        # if the pipeline is paused we record the tasks to be sent into the pipeline state
        if  context.pipeline_state.status.value == TaskStatus.PAUSED:
            
            saved_to_store = False

            if self._storage is not None and context.pipeline_state_size >= self._storage.threshold:
                proto = PausedTask(destination=destination, task_request=task_request)
                saved_to_store = await self._try_save_to_store(context, proto, task_request.uid)
                            
            if not saved_to_store:
                paused_task = PausedTask(destination=destination, task_request=task_request)
            else:
                paused_task = PausedTask(destination=destination, task_uid=task_request.uid)
            
            context.pipeline_state.paused_tasks.append(paused_task)

        # else send the tasks to their destinations
        else:
            context.send_message(destination, task_request.id, task_request)

    async def _try_save_to_store(self, context, paused_task, task_uid):
        try:
            keys = [context.pipeline_state.invocation_id, task_uid]
            await self._storage.store(keys, paused_task.SerializeToString())
            return True
        except Exception as ex:
            _log.warning(f'Error saving {keys} to backend storage - {ex}')
            return False

    async def _load_from_store(self, context, paused_task):
        keys = ['paused_tasks', context.pipeline_state.invocation_id, paused_task.task_uid]

        proto_bytes = await self._storage.fetch(keys)
        paused_task = PausedTask()
        paused_task.ParseFromString(proto_bytes)
        return paused_task



from unittest.case import skip
from statefun_tasks.context import TaskContext
from statefun_tasks.pipeline_impl.handlers import PipelineMessageHandler
from statefun_tasks.types import Task, Group, TasksException
from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TaskException, TaskStatus
from typing import Union


class ContinuePipelineHandler(PipelineMessageHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def can_handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException]) -> bool:
        return context.pipeline_state is not None \
            and context.pipeline_state.invocation_id == message.invocation_id \
                and context.pipeline_state.status.value in [TaskStatus.RUNNING, TaskStatus.PAUSED] \
                    and isinstance(message, (TaskResult, TaskException))

    async def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline, **kwargs):
        task_result_or_exception = message
        caller_uid = message.uid

        # mark pipeline step as complete
        self.graph.mark_task_complete(caller_uid, task_result_or_exception)

        # notify event handler (with option to cancel)
        try:
            pipeline.events.notify_pipeline_task_finished(context, task_result_or_exception)
        except TasksException as ex:
            await pipeline.cancel(context, ex)
            return False, task_result_or_exception

        # release deferred tasks if they can now run
        await self.submitter.release_tasks(context, caller_uid, task_result_or_exception)

        # get the next step of the pipeline to run (if any)
        current_step, next_step, group, empty_group = self.graph.get_next_step_in_pipeline(caller_uid)

        # if this task is part group then we need to record the results so we can aggregate later
        if group is not None:
            await self.result_aggregator.add_result(context, caller_uid, task_result_or_exception)

            # once the group is complete aggregate the results
            if group.is_complete():
                task_result_or_exception = await self.result_aggregator.aggregate(context, group)

                # pause the pipeline if this completed group is a wait
                if group.is_wait:
                    await pipeline.pause(context)
        else:
            # pause the pipeline if this task is a wait
            if current_step.is_wait:  
                await pipeline.pause(context)

        # if we got an exception then the next step is the finally_task if there is one (or none otherwise)
        if isinstance(task_result_or_exception, TaskException):
            next_step = self.graph.try_get_finally_task(caller_uid)

        # else if we came across an empty group between this task and next_entry then our result must be an empty array []
        # as we cannot call an empty group but can synthesise the result (remembering to pass through state)
        elif empty_group:
            _, state = self._serialiser.deserialise_result(task_result_or_exception)
            self._serialiser.serialise_result(task_result_or_exception, ([]), state)

        # turn next step into remainder of tasks to call
        if isinstance(next_step, Task):
            tasks = [next_step]
            max_parallelism = 1

            if next_step.is_finally:
                # record the result of the task prior to the finally task so we can return it once the finally task completes
                pipeline.save_result_before_finally(context, task_result_or_exception)
            
        elif isinstance(next_step, Group):
            tasks, max_parallelism, _ = self.graph.get_initial_tasks(group=next_step)
        else:
            tasks = []

        if any(tasks):
            # split into tasks to call now and those to defer if max parallelism is exceeded
            await self.submitter.submit_tasks(context, tasks, task_result_or_exception=task_result_or_exception, max_parallelism=max_parallelism)

        else:
            last_step = self._pipeline[-1]

            if last_step.is_complete():
                # if we are at the last step in the pipeline then we are complete
                context.pipeline_state.status.value = TaskStatus.COMPLETED if isinstance(task_result_or_exception, TaskResult) else TaskStatus.FAILED
                pipeline.events.notify_pipeline_status_changed(context, context.pipeline_state.pipeline, context.pipeline_state.status.value)


            elif isinstance(task_result_or_exception, TaskException):
                if group is None or group.is_complete():
                    # else if have an exception then we failed but waiting for any parallel tasks in the group to complete first
                    context.pipeline_state.status.value = TaskStatus.FAILED
                    pipeline.events.notify_pipeline_status_changed(context, context.pipeline_state.pipeline, context.pipeline_state.status.value)

        # continue
        return True, task_result_or_exception
